/**
  *  \file client/scr-base.cc
  *  \brief Starbase Screen
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "client/cscreen.h"
#include "client/marks.h"
#include "game/global.h"
#include "gfx/font.h"
#include "gfx/gfx.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "ui/msgbox.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "client/actions.h"
#include "game/ship.h"
#include "ui/listbox.h"
#include "ui/widgets.h"
#include "client/dlg-tech.h"
#include "ui/table.h"
#include "client/dlg-structure.h"
#include "client/dlg-ammo.h"
#include "client/dlg-build.h"
#include "resmgr/resid.h"
#include "game/tables.h"
#include "client/wscanchart.h"
#include "client/dlg-comment.h"
#include "int/intglobal.h"
#include "game/config.h"
#include "game/unitscore.h"
#include "client/tiles/planetfcode.h"
#include "client/tiles/planetbase.h"
#include "client/tiles/baseorder.h"
#include "client/tiles/basetech.h"
#include "client/tiles/basemineral.h"
#include "client/dlg-objsel.h"
#include "client/tiles/baseoverview.h"

namespace {
    class WBaseScreen : public WControlScreen {
     public:
        WBaseScreen();
        ~WBaseScreen();
        bool handleEvent(const UIEvent& e, bool second_pass);
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void update();
    };

    class WBaseScreenHeaderTile : public WControlScreenHeaderTile {
     public:
        WBaseScreenHeaderTile(GObjectType& type);
        string_t getTitle();
        string_t getSubtitle();
        ResId getPictureId();
    };
}

int
chooseBase()
{
    int nbases = global_base_type.getCount();
    if (nbases == 0) {
        messageBox(_("You do not have any starbases. "
                     "To build starbases, use [F8] on a planet."),
                   _("Starbases"));
        return 0;
    } else if (nbases == 1) {
        return global_base_type.getCurrent();
    } else {
        GAliasObjectType aot(global_base_type);
        WBaseOverviewTile bot(aot);
        WObjectSelectionDialog dlg(_("Select Starbase"), aot);
        dlg.addWidget(bot);
        dlg.pack();
        ui_root->centerWidget(dlg);
        if (ui_root->runModal(dlg) == cm_Enter)
            return aot.getCurrent();
        else
            return 0;
    }
}

/****************************** WBaseScreen ******************************/

WBaseScreen::WBaseScreen()
    : WControlScreen(global_base_type, WScreenId::BaseScreen)
{
    addTile(holder.add(new WBaseScreenHeaderTile(global_base_type)));
    addTile(holder.add(new WBaseMineralTile(global_base_type)), _("Minerals & Funds:"));
    addTile(holder.add(new WBaseTechTile(global_base_type)),    _("Technology & Defense:"));
    addTile(holder.add(new WBaseOrderTile(global_base_type)),   _("Orders:"));
    addTile(holder.add(new WPlanetFCodeTile(global_base_type)));
    addTile(holder.add(new WPlanetBaseTile(global_base_type, true)));
    addChart(holder.add(new WScannerChartWidget()));
}

WBaseScreen::~WBaseScreen()
{ }

static void
doBaseXchg(WControlScreen& scr)
{
    WScanner& scanner = scr.getScanner();
    int pid = scanner.getScannedPlanetId(scanner.getScannedShipId() != 0);
    if (pid && pid != scr.getType().getCurrent()) {
        if (scr.getType().isExistant(pid))
            scr.stopWithScreenId(WScreenId(WScreenId::BaseScreen, pid));
        else if (global_planet_type.isExistant(pid))
            scr.stopWithScreenId(WScreenId(WScreenId::PlanetScreen, pid));
        else
            putKey(ss_Ctrl + SDLK_F1);
    } else {
        putKey(ss_Ctrl + SDLK_F1);
    }
}

bool
WBaseScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_Base_Recycle) {
        int pid         = getType().getCurrent();
        GUniverse& univ = *getType().getUniverse();
        GPlanet& me     = univ.getPlanet(pid);
        doBaseFixRecycle(univ, me, sender);
        return true;
    }
    return WControlScreen::handleCommand(cmd, sender);
}

bool
WBaseScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        try {
            int pid         = getType().getCurrent();
            GUniverse& univ = *getType().getUniverse();
            GPlanet& me     = univ.getPlanet(pid);
            switch (e.key.code) {
             case 'x':
                doBaseXchg(*this);
                return true;
             case 'f':
                doPlanetFCodeChange(univ, me);
                return true;
             case 'c':
             case 'u':
                doPlanetTransfer(univ, me, e.key.code == 'u');
                return true;
             case SDLK_F8:
                stopWithScreenId(WScreenId(WScreenId::PlanetScreen, pid));
                return true;
             case 'm':
                doBaseChangeMission(univ, me);
                return true;
             case ss_Ctrl + 'r':
                if (int sid = me.getBaseShipyardId())
                    stopWithScreenId(WScreenId(WScreenId::ShipScreen, sid));
                return true;
             case ss_Ctrl + 'c':
                if (int sid = me.getShipCloningHere(univ))
                    stopWithScreenId(WScreenId(WScreenId::ShipScreen, sid));
                return true;
             case 't':
                doTechUpgradeDialog(univ, me);
                return true;
             case 'd':
                doPlanetBuildScreen(univ, me, 2);
                return true;
             case 'a':
                doBaseBuildAmmo(univ, me);
                return true;
             case 'b':
             case SDLK_TAB:
             {
                 WShipBuildScreenWithActor bs(me);
                 ui_root->centerWidget(bs);
                 ui_root->add(bs);
                 ui_root->run();
                 ui_root->close(bs);
                 univ.doScreenUpdates();
                 return true;
             }
             case SDLK_F9:
                doEditComment(me, getPlanetProperties(me.getId()), _("Enter new comment for this planet:"), _("Edit Comment"));
                return true;
             case SDLK_INSERT:
                doPlanetAddToSim(me);
                return true;
             /* base screen keys:
                's'    ship cost
                alt-r  reset
                f10    choose fleet */

            }
        }
        catch (const GError& e) {
            messageBox(e.getUserError(), _("Error"));
            return true;
        }
    }
    return WControlScreen::handleEvent(e, second_pass);
}

void
WBaseScreen::update()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(&type.getObject(type.getCurrent())))
        getScanner().getChart()->setCenterPosition(mo->getPos());
}

WBaseScreenHeaderTile::WBaseScreenHeaderTile(GObjectType& type)
    : WControlScreenHeaderTile(type, bAuto + bAdd)
{ }

string_t
WBaseScreenHeaderTile::getTitle()
{
    return getObject()->getName();
}

string_t
WBaseScreenHeaderTile::getSubtitle()
{
    string_t result;
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        int16 level = p->unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpLevel));
        const char* fmt = (level >= 0
                           ? _("(Id #%d, %s)")
                           : _("(Id #%d)"));
        result = format(fmt, getObject()->getId(), getExperienceLevelName(level));
    }
    return result;
}

ResId
WBaseScreenHeaderTile::getPictureId()
{
    int max_tech = 0;
    int id = 0;
    if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
        id = p->getId();
        if (p->hasBase()) {
            max_tech = std::max(std::max(p->getBaseTech(tl_Engine),
                                         p->getBaseTech(tl_Hull)),
                                std::max(p->getBaseTech(tl_Beam),
                                         p->getBaseTech(tl_Torpedo)));
        }
    }
    return ResId(res::base, max_tech, id);
}

WScreenId
doBaseScreen()
{
    return WBaseScreen().runScreen();
}
