/**
  *  \file client/tiles/vishull.cc
  *  \brief Visual Scanner: Hull Information Tile
  */

#include "client/tiles/vishull.h"
#include "cpluslib/format.h"
#include "game/ship.h"
#include "ui/uiglobal.h"
#include "util/nls.h"

/************************ WVisualScannHullInfoTile ***********************/

WVisualScanHullInfoTile::WVisualScanHullInfoTile(GObjectType& type)
    : WObjectTypeObserverTile(font_sizes[FONT_NORMAL].scaledBy(10, 6), type)
{ }

void
WVisualScanHullInfoTile::drawData(GfxCanvas& can)
{
    setColor(UIColor::tc_Static);

    GShip* s;
    int hid;
    if ((s = dynamic_cast<GShip*>(getObject())) != 0 && (hid = s->getHullId()) > 0) {
        const GHull& h = getHull(hid);
        const int w = extent.w;
        const int hh = font_heights[FONT_NORMAL];

        // Line 1: "Hull mass: nn kt"
        string_t text = format(_("Hull mass: %d kt"), h.getMass());
        outTextF(can, FONT_NORMAL, 0, 0, w, text);

        // Line 2:
        //   Cargo: a/b kt
        //   Max cargo: x kt
        if (s->getKind() == GShip::CurrentShip)
            text = format(_("Cargo: %d/%d kt"), h.getCargo() - s->getFreeCargo(), h.getCargo());
        else
            text = format(_("Max Cargo: %d kt"), h.getCargo());
        outTextF(can, FONT_NORMAL, 0, hh, w, text);

        // Line 3:
        //   Fuel: a/b kt
        //   Max fuel: x kt
        if (s->getKind() == GShip::CurrentShip)
            text = format(_("Fuel: %d/%d kt"), s->getCargo(el_Neutronium), h.getFuel());
        else
            text = format(_("Max Fuel: %d kt"), h.getFuel());
        outTextF(can, FONT_NORMAL, 0, 2*hh, w, text);

        // Line 4:
        //   3xBeamTyp
        //   Max Beams: x
        //   No beams
        // FIXME: handle history
        if (s->getKind() == GShip::CurrentShip) {
            if (s->getBeamCount() > 0)
                text = format("%dx%s", s->getBeamCount(), getBeam(s->getBeamType()).getName());
            else
                text = _("No beams");
        } else {
            if (h.getMaxBeams() > 0)
                text = format(_("Max Beams: %d"), h.getMaxBeams());
            else
                text = _("No beams");
        }
        outTextF(can, FONT_NORMAL, 0, 3*hh, w, text);

        // Line 5:
        //   2xTorpTyp
        //   Max torps: x
        //   No torps
        //   Fighter bays: x
        //   No fighter bays
        if (s->getKind() == GShip::CurrentShip) {
            if (s->getTorpLauncherCount() > 0) {
                text = format("%dx%s", s->getTorpLauncherCount(), getTorp(s->getTorpType()).getName());
            } else if (s->getBayCount() > 0) {
                text = format(_("Fighter bays: %d"), s->getBayCount());
            } else if (h.getNumBays() > 0) {
                text = _("No fighter bays");
            } else if (h.getMaxLaunchers() > 0) {
                text = _("No torps");
            } else {
                text.clear();
            }
        } else {
            if (h.getNumBays() > 0)
                text = format(_("Fighter bays: %d"), h.getNumBays());
            else if (h.getMaxLaunchers() > 0)
                text = format(_("Max torps: %d"), h.getMaxLaunchers());
            else
                text.clear();
        }
        outTextF(can, FONT_NORMAL, 0, 4*hh, w, text);

        // Line 6:
        //   FCode: foo
        if (s->getKind() == GShip::CurrentShip)
            text = format(_("FCode: %s"), s->getFCode());
        else
            text.clear();
        outTextF(can, FONT_NORMAL, 0, 5*hh, w, text);
    } else {
        // Hull not known: completely blank
        getSkin().drawBackground(can, getExtent());
    }
}

void
WVisualScanHullInfoTile::getLayoutInfo(LayoutInfo& info)
{
    WObjectTypeObserverTile::getLayoutInfo(info);
    info.flags |= li_FlexibleH;
}

