/**
  *  \file game/cargo.cc
  *  \brief Cargo Reservation/Arbitration Primitives
  *
  *  PCC II supports "concurrent" access to cargo containers. Hence,
  *  we use a transaction-like scheme to ensure correct behaviour. 
  *  Each object has a cargo arbiter (GCargoArbiter), and a number of cargo
  *  traits (GCargoContainer). Cargo which is removed from the object
  *  is first reserved, and not added to the target object until the
  *  transaction is committed. We don't support "real", preemptive
  *  concurrency, though.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/cargo.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "game/spec.h"

/***************************** GCargoArbiter *****************************/

struct GCargoArbiter::Impl {
    /** Reference count. Equals the number of incUseCount() calls
        which have not been taken back using decUseCount() yet. */
    int ref_count;
    /** Reserved cargo which is not available. */
    int32 reserved_cargo[el_COUNT];

    Impl();
};

/** Constructor. */
GCargoArbiter::Impl::Impl()
    : ref_count(0)
{
    std::fill_n(reserved_cargo, int(el_COUNT), 0);
}

GCargoArbiter::~GCargoArbiter() throw()
{
    delete data;
}

/** Increment use count. Every piece of code which possibly does a
    cargo reservation must call this. */
void
GCargoArbiter::incUseCount()
{
    if (!data)
        data = new Impl();
    ++data->ref_count;
}

/** Decrement use count. You may no longer call changeReservedCargo()
    after calling decUseCount(). */
void
GCargoArbiter::decUseCount() throw()
{
    if (!--data->ref_count) {
        delete data;
        data = 0;
    }
}

/** Get reserved cargo. User interface should not include this amount
    in display. Every transaction must assume that this amount is not
    available (except for cargo known to reserved by the transaction
    itself). */
int32
GCargoArbiter::getReservedCargo(GCargoType t) const throw()
{
    return data ? data->reserved_cargo[t] : 0;
}

/** Change cargo reservation.
    \param t cargo type
    \param amount >0 to reserve cargo, <0 to give back
    \returns new amount of reserved cargo. */
int32
GCargoArbiter::changeReservedCargo(GCargoType t, int32 amount) throw()
{
    ASSERT(data);
    ASSERT(data->reserved_cargo[t] + amount >= 0);
    data->reserved_cargo[t] += amount;
    return data->reserved_cargo[t];
}

/**************************** GCargoContainer ****************************/

/*! \class GCargoContainer

    This is the base class for everything that deals with cargo
    (minerals, sup, mc, fighters, torps, clans). It is used for cargo
    transfer and some other things.

    Derived classes provide access to the underlying object.
    - getCargoOnObject() yields the current amount on the object, where
      reserved stuff has already been subtracted (that is, the
      amount which can still be safely subtracted).
    - changeCargoOnObject() changes the current physical amount.
    - changeReservedCargo() changes the current reserved amount.
    - getMaxCargo() yields the maximum amount which can appear on the
      object.

    The public manipulation interface uses these functions to provide
    safe cargo amount manipulation. Using commit(), you can commit the
    changes; cancel() takes them back.

    \todo FIXME: a problem which can not be solved is to test whether a
    package of multiple cargo types can fit onto a container without
    modifying it, e.g. to test whether I can add "50TDM 2800$". Things
    I want to do with this:
    - "load minerals for as many torps as possible" (mkt support)
    - clearing the hold space in multi-cargo transfer.

    It is still possibly doing this item-by-item, but that is
    cumbersome.

    Interfaces that come to mind (public virtual functions):
    - bool canAdd(GCargoContainer& other): return true iff other
      can be added to *this. This solves problem 2; problem 1 can
      be solved using binary intersection tests.
    - int divide(GCargoContainer& other): how many times other
      can be added to *this. Elegant at first, this one has problems
      handling "infinite" containers (planets). */

/** Constructor. Creates a cargo container which did not yet
    add/remove anything from the underlying object. */
GCargoContainer::GCargoContainer()
    : flags(0)
{
    std::fill_n(diffs, countof(diffs), 0);
}

GCargoContainer::~GCargoContainer()
{ }

/** Return current amount of cargo. This is the value to be presented
    to the user. It includes others' removals but not their additions,
    and it includes our removals and additions. Therefore it is safe
    in the sense that the value returned from this function can be
    removed from the object without problems. */
int32
GCargoContainer::getCargo(GCargoType type)
{
    if (diffs[type] < 0) {
        /* we have removed some cargo from this object. Hence, getAmount()
           will include our reservation. */
        return getCargoOnObject(type);
    } else {
        /* we are receiving cargo. getAmount() will include others'
           reservations. */
        return getCargoOnObject(type) + diffs[type];
    }
}

/** Remove some cargo from this object. This will not actually remove
    the cargo from the underlying object, it will merely reserve it
    for us.
    \param type     cargo type
    \param amount   amount to remove
    \param partial  allow partial removal

    \return amount removed. If partial=false, either 0 or /amount/. 
    Otherwise, any value from [0, amount]. */
int32
GCargoContainer::removeCargo(const GCargoType type, int32 amount, const bool partial)
{
    /* quick exit */
    if (amount == 0)
        return 0;

    ASSERT(amount >= 0);
    ASSERT(!isFinished());

    const int32_t have = getCargo(type);
    if (amount > have) {
        // We're trying to remove more than we have. Exit without any
        // change if we're not permitted to do partial change, or if
        // we don't have anything.
        if (!partial || have == 0)
            return 0;
        amount = have;
    }

    // Cargo removal is split into undoing a portion of previously-received
    // cargo, and reservation of new cargo.
    const int32_t previouslyReceived = std::max(std::min(diffs[type], amount), int32_t(0));
    const int32_t newlyReserved = amount - previouslyReceived;

    // Now reserve the remainder
    changeReservedCargo(type, newlyReserved);
    diffs[type] -= amount;
    sig_changed.raise();

    return amount;
}

/** Add cargo to this object. This will not actually add the cargo to
    the object, but will just remember to do that at a later time.
    \param type    cargo type
    \param amount  amount to add
    \param partial allow partial add

    \return amount added. If partial=false, either 0 or /amount/. 
    Otherwise, any value from [0,amount]. */
int32
GCargoContainer::addCargo(const GCargoType type, int32 amount, const bool partial)
{
    if (amount == 0)
        return 0;

    ASSERT(amount >= 0);
    ASSERT(!isFinished());

    /* add some cargo. First check whether we can actually have that much. */
    const int32_t space = getMaxCargo(type) - getCargo(type);
    if (space <= 0) {
        /* we are already overloaded */
        return 0;
    }
    if (amount > space) {
        if (!partial)
            return 0;
        amount = space;
    }

    /* Cargo addition is split into undoing a previous reservation and
       adding new cargo. */
    const int32_t previouslyReserved = std::max(std::min(-diffs[type], amount), int32_t(0));
    // const int32_t newlyReceived = amount - previouslyReserved;

    changeReservedCargo(type, -previouslyReserved);
    diffs[type] += amount;
    sig_changed.raise();

    return amount;
}

/** Add or remove cargo. Same as addCargo() if amount is positive,
    same as removeCargo() if amount is negative. See there for
    details. */
int32
GCargoContainer::addOrRemoveCargo(const GCargoType type, const int32 amount, const bool partial)
{
    if (amount > 0)
        return addCargo(type, amount, partial);
    else if (amount < 0)
        return -removeCargo(type, -amount, partial);
    else
        return 0;
}

/** Cancel transaction. This will undo our reservations. */
void
GCargoContainer::cancel()
{
    /* cancelling a transaction means canceling our reservations */
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i) {
        if (diffs[i] < 0) {
            changeReservedCargo(i, -diffs[i]);
            diffs[i] = 0;
        }
    }
}

/** Commit transaction. This will perform all changes. */
void
GCargoContainer::commit()
{
    /* commit means to cancel our reservations *and* apply our changes */
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i) {
        if (diffs[i] < 0)
            changeReservedCargo(i, -diffs[i]);
        if (diffs[i] != 0)
            changeCargoOnObject(i, diffs[i]);
        diffs[i] = 0;
    }
}
